/* oopdefs.c */

#include <stdlib.h>
#include <stdio.h>
#include "opldefs.h"


/******************************************************************************
 * Lists
 */

procedureList *new_procedureList( procedure *el, procedureList *l )
{
  procedureList *ret = (procedureList*) malloc( sizeof( procedureList ) );
  ret->head = el;
  ret->tail = l;
  return ret;
}


identList *new_identList( ident *el, identList *l )
{
  identList *ret = (identList*) malloc( sizeof( identList ) );
  ret->head = el;
  ret->tail = l;
  return ret;
}


statementList *new_statementList( statement *el, statementList *l )
{
  statementList *ret = (statementList*) malloc( sizeof( statementList ) );
  ret->head = el;
  ret->tail = l;
  return ret;
}


expressionList *new_expressionList( expression *el, expressionList *l )
{
  expressionList *ret = ( expressionList *) malloc( sizeof( expressionList ) );
  ret->head = el;
  ret->tail = l;
  return ret;
}


/******************************************************************************
 * Identifiers, Procedures and Programs
 */

ident* new_ident( char* n, identType t, struct expressionList *pa )
{
  ident* ret = (ident*) malloc( sizeof( ident ) );
  ret->name    = n;
  ret->type    = t;
  ret->pArrays = pa;
  return ret;
}


procedure* new_procedure( ident* n, identList* pa, struct statementList *ps )
{
  procedure* ret = (procedure*) malloc( sizeof( procedure ) );
  ret->name   = n;
  ret->pArgs  = pa;
  ret->pStmts = ps;
  return ret;
}


program* new_program( procedureList *pp )
{
  program* ret = (program*) malloc( sizeof( program ) );
  ret->pProcs = pp;
  ret->sourcefile = NULL;
  ret->outputfile = NULL;
  return ret;
}


/******************************************************************************
 * Statements
 */


statement* new_if_statement( struct expression *c,
  struct statementList *pt, struct statementList *pe )
{
  statement* ret = (statement*) malloc( sizeof( statement ) );
  ret->kind = if_stmt;
  ret->data.if_part.condition = c;
  ret->data.if_part.pThen = pt;
  ret->data.if_part.pElse = pe;
  return ret;
}


statement* new_while_statement( struct expression *c, struct statementList *pb )
{
  statement* ret = (statement*) malloc( sizeof( statement ) );
  ret->kind = while_stmt;
  ret->data.while_and_do_part.condition = c;
  ret->data.while_and_do_part.pBody = pb;
  return ret;
}


statement* new_do_statement( struct expression *c, struct statementList *pb )
{
  statement* ret = (statement*) malloc( sizeof( statement ) );
  ret->kind = do_stmt;
  ret->data.while_and_do_part.condition = c;
  ret->data.while_and_do_part.pBody = pb;
  return ret;
}


statement* new_assignment_statement( ident *n, struct expression *v )
{
  statement* ret = (statement*) malloc( sizeof( statement ) );
  ret->kind = assignment_stmt;
  ret->data.assignment_part.name = n;
  ret->data.assignment_part.value = v;
  return ret;
}


statement* new_global_statement( identList *pi )
{
  statement* ret = (statement*) malloc( sizeof( statement ) );
  ret->kind = global_stmt;
  ret->data.global_and_local_part.pVars = pi;
  return ret;
}


statement* new_local_statement( identList *pi )
{
  statement* ret = (statement*) malloc( sizeof( statement ) );
  ret->kind = local_stmt;
  ret->data.global_and_local_part.pVars = pi;
  return ret;
}


statement* new_call_statement( ident *n, struct expressionList *pa, int opl )
{
  statement* ret = (statement*) malloc( sizeof( statement ) );
  ret->kind = call_stmt;
  ret->data.call_part.opl_call = opl;
  ret->data.call_part.name =n;
  ret->data.call_part.pArgs = pa;
  return ret;
}


statement* new_return_statement( struct expression *e )
{
  statement* ret = (statement*) malloc( sizeof( statement ) );
  ret->kind = return_stmt;
  ret->data.return_part.expr = e;
  return ret;
}


/******************************************************************************
 * Expressions
 */

expression* new_variable_expr( ident *v )
{
    expression* ret = (expression*) malloc( sizeof( expression ) );
    ret->kind               = variable_expr;
	ret->type               = v->type;
    ret->data.variable_part = v;
	return ret;
}

expression* new_integer_expr( int k, int n )
{
	expression* ret = (expression*) malloc( sizeof( expression ) );
	ret->kind              = k;
	ret->type              = integer_type;
	ret->data.integer_part = n;
	return ret;
}

/*expression* new_floating_point_expr(); - not needed? */
/*expression* new_long_integer_expr(); - not needed? */

expression* new_string_expr( const char *s )
{
    expression* ret = (expression*) malloc( sizeof( expression ) );
    ret->kind              = string_expr;
	ret->type              = string_type;
    ret->data.string_part  = s;
	return ret;
}


expression* new_binary_op_expression( exprKind k, expression* l, expression* r )
{
  expression* ret = (expression*) malloc( sizeof( expression ) );
  ret->kind = k;
  ret->data.binary_op_part.left = l;
  ret->data.binary_op_part.right = r;

	/* expression type checking code:
     *
     * addition, subtraction, multiplication, division = type of params
     * and, or = boolean or integer depending in context!
     * equality, inequality, lessthan, greaterthan, lessequal, greaterequal = boolean
     *
     * parameter types work on simple promotion and checks against conversion
     * of strings.
	 */

	switch( k )
	{
	case and_expr: case or_expr:
		if( (l->type == integer_type) && (r->type == integer_type) )
		  ret->type = integer_type;
		else
		  ret->type = boolean_type;
		return ret;
    case equality_expr: case inequality_expr:
    case lessthan_expr: case greaterthan_expr:
	case lessequal_expr: case greaterequal_expr:
		ret->type = boolean_type;
		return ret;
	}

	if( l->type == r->type )
	{
		ret->type = l->type;
		return ret;
	}

	switch( l->type )
	{
	case integer_type:
		switch( r->type )
		{
		case float_type:   ret->type = float_type; return ret;
		case string_type:  printf( "ERROR: Can't convert string to integer\n" ); break;
		case long_type:    ret->type = long_type; return ret;
        case boolean_type: printf( "WARNING: Conversion of integer to boolean\n" ); ret->type = boolean_type; return ret;
		}
		break;
	case float_type:
		switch( r->type )
		{
		case integer_type: ret->type = float_type; return ret;
		case string_type:  printf( "ERROR: Can't convert string to float\n" ); break;
		case long_type:    ret->type = float_type; return ret;
        case boolean_type: printf( "ERROR: Can't convert float to boolean\n" ); break;
		}
		break;
	case string_type:
		switch( r->type )
		{
		case integer_type: printf( "ERROR: Can't convert string to integer\n" ); break;
		case float_type:   printf( "ERROR: Can't convert string to float\n" ); break;
		case long_type:    printf( "ERROR: Can't convert string to long\n" ); break;
        case boolean_type: printf( "ERROR: Can't convert string to boolean\n" ); break;
		}
		break;
	case long_type:
		switch( r->type )
		{
		case integer_type: ret->type = long_type; return ret;
		case float_type:   ret->type = float_type; return ret;
		case string_type:  printf( "ERROR: Can't convert string to long\n" ); break;
        case boolean_type: printf( "WARNING: Conversion of long to boolean\n" ); ret->type = boolean_type; return ret;
		}
		break;
    case boolean_type:
		switch( r->type )
		{
		case integer_type: printf( "WARNING: Conversion of integer to boolean\n" ); ret->type = boolean_type; return ret;
		case float_type:   printf( "ERROR: Can't convert float to boolean\n" ); break;
		case string_type:  printf( "ERROR: Can't convert string to boolean\n" ); break;
		case long_type:    printf( "WARNING: Conversion of long to boolean\n" ); ret->type = boolean_type; return ret;
		}
		break;
	}

	ret->type = untyped_id;
    return ret;
}

expression* new_unary_op_expression( exprKind k, expression* e )
{
  expression* ret = (expression*) malloc( sizeof( expression ) );
  ret->kind = k;
  ret->type = e->type;
  ret->data.unary_op_part.expr = e;
  return ret;
}

expression* new_call_expression( ident* n, expressionList* pa, int opl )
{
  expression* ret = (expression*) malloc( sizeof( expression ) );
  ret->kind = call_expr;
  ret->type = n->type;
  ret->data.call_part.opl_call = opl;
  ret->data.call_part.name = n;
  ret->data.call_part.pArgs = pa;
  return ret;
}


/******************************************************************************
 * Utility
 */

const char* typeToString( exprType t )
{
	switch( t )
	{
	case integer_type: return "integer";
    case float_type:   return "float";
	case string_type:  return "string";
    case long_type:    return "long";
	case boolean_type: return "boolean";
	case untyped_type:
	default:
		return "<untyped>";
	}
}